"use strict";

/**
 * Defines the known stroke styles and methods to draw styled outer lines.
 */
function LineStrokes()
{
   /**
    * Joins path segments by extending their outside edges until
    * they meet.
    */
   var JOIN_MITER = 0;

   /**
    * Joins path segments by rounding off the corner at a radius
    * of half the line width.
    */
   var JOIN_ROUND = 1;

   /**
    * Joins path segments by connecting the outer corners of their
    * wide outlines with a straight segment.
    */
   var JOIN_BEVEL = 2;

   /**
    * Ends unclosed subpaths and dash segments with no added
    * decoration.
    */
   var CAP_BUTT = 0;

   /**
    * Ends unclosed subpaths and dash segments with a round
    * decoration that has a radius equal to half of the width
    * of the pen.
    */
   var CAP_ROUND = 1;

   /**
    * Ends unclosed subpaths and dash segments with a square
    * projection that extends beyond the end of the segment
    * to a distance equal to half of the line width.
    */
   var CAP_SQUARE = 2;
   
   /**
    * This is the list of all possible line stoke types.
    */
   var LineStrokeEnum = 
   {
      DEFAULT           : 0,
      SPECIFIED_WIDTH   : 1,
      DOTTED            : 2,
      DOTTED_SMALL      : 3,
      DOTTED_SMALL_THIN : 4
   };

   var LineStroke =
   {
      /** Dotted line specific parameter. */
      DOT_STEP            : 2.0,
      
      /** Dotted line specific parameter. */
      DOT_STEP_SMALL      : 1.0,
      
      /** Dotted line width. */
      DOT_LINE_WIDTH      : 1.0,
      
      /** Dotted thin line width. */
      DOT_LINE_THIN_WIDTH : 0.5
   };
   
   var dotStroke = new BasicStroke(LineStroke.DOT_LINE_WIDTH, CAP_BUTT, JOIN_BEVEL, 0.0,
         [LineStroke.DOT_STEP, LineStroke.DOT_STEP], LineStroke.DOT_STEP);

   var dotSmallStroke  = new BasicStroke(LineStroke.DOT_LINE_WIDTH, CAP_BUTT, JOIN_BEVEL, 0.0,
         [LineStroke.DOT_STEP_SMALL, LineStroke.DOT_STEP_SMALL], LineStroke.DOT_STEP_SMALL);

   var dotSmallThinStroke = new BasicStroke(LineStroke.DOT_LINE_THIN_WIDTH, CAP_BUTT,
         JOIN_BEVEL, 0.0, [LineStroke.DOT_STEP_SMALL, LineStroke.DOT_STEP_SMALL],
         LineStroke.DOT_STEP_SMALL);

   var defaultStroke = new BasicStroke(1.0, CAP_SQUARE, JOIN_MITER, 10.0, null, 0.0);
   
   /** Holds all java script styles of how the ends of lines are drawn. */
   var caps = [ "butt", "round", "square"];
   
   /**
    *   Holds all java script styles of how two connecting segments with non-zero lengths
    *   in a shape are joined together.
    */
   var joins = [ "miter", "round", "bevel"];
   
   /**
    * Holds properties for line drawing styles.
    * 
    * @param    width
    *           The width of lines to draw.
    * @param    cap
    *           The decoration of the line ends.
    * @param    join
    *           The decoration applied where path segments meet.
    * @param    miterlimit
    *           The limit to trim the miter join.
    * @param    dash
    *           The array representing the dashing pattern.
    * @param    dash_phase
    *           The offset to start the dashing pattern.
    */
   function BasicStroke(width, cap, join, miterLimit, dash, dash_phase)
   {
      this.getWidth = function() { return width; };
      this.getCap   = function() { return caps[cap]; };
      
      this.getJoin        = function() { return joins[join]; };
      this.getMiterLimit  = function() { return miterLimit; };
      
      this.getDash       = function() { return dash; };
      this.getDashPhase  = function() { return dash_phase; };
   };

   /**
    * Returns the stroke style object.
    * 
    * @param    {Number} strokeStyle
    *           The stroke style id according to the LineStroke enumeration.
    * @param    {Number} width
    *           The line width.
    * @return   The basic line stroke object.
    */
   function getLineStroke(strokeStyle, width)
   {
      switch(strokeStyle)
      {
         case LineStrokeEnum.SPECIFIED_WIDTH:
            return new BasicStroke(width, CAP_SQUARE, JOIN_MITER, 10.0, null, 0.0);
         case LineStrokeEnum.DOTTED:
            return dotStroke;
         case LineStrokeEnum.DOTTED_SMALL:
            return dotSmallStroke;
         case LineStrokeEnum.DOTTED_SMALL_THIN:
            return dotSmallThinStroke;
         default:
            return defaultStroke;
      }
   };
   
   /**
    * Apply the given stroke style to the JS native canvas renderer.
    * 
    * @param    {CanvasRenderingContext2D} context
    *           The canvas renderer.
    * @param    {Number} strokeStyle
    *           The stroke style id.
    * @param    {Number} width
    *           The line width.
    */
   this.apply = function(context, strokeStyle, width)
   {
      if (context instanceof CanvasRenderingContext2D)
      {
         var basicStroke = getLineStroke(strokeStyle, width);
         context.lineWidth = basicStroke.getWidth();
         
         context.lineCap = basicStroke.getCap();
         
         context.lineJoin = basicStroke.getJoin();
         
         context.miterLimit = basicStroke.getMiterLimit();
         if (basicStroke.getDash())
         {
            context.setLineDash(basicStroke.getDash());
         }
         else
         {
            context.setLineDash([]);
         }
         context.lineDashOffset = basicStroke.getDashPhase();
      }
   }

   /**
    * Apply the given stroke style to the area outer bounds.
    * 
    * @param    {CanvasRenderingContext2D} context
    *           The canvas renderer.
    * @param    {Number} strokeStyle
    *           The stroke style id.
    * @param    {Number} width
    *           The line width.
    * @param    {Number[]} strokeColor
    *           The array of 3 integer values between 0 and 255 inclusive, representing
    *           a stroke color.
    * @param    {Number[]} fillColor
    *           The array of 3 integer values between 0 and 255 inclusive, representing
    *           a fill color.
    * @param    {Array} path
    *           The array of points that forms the outline of the target area.
    */
   this.applyStrokeToPath = function(context, strokeStyle, width, strokeColor, fillColor, path)
   {
      if (context instanceof CanvasRenderingContext2D)
      {
         var basicStroke = getLineStroke(strokeStyle, width);
         var renderer;
         if (strokeStyle == LineStrokeEnum.SPECIFIED_WIDTH)
         {
            renderer =  new WidenPathRenderer(context, basicStroke, strokeColor, fillColor);
         }
         else
         {
            renderer =  new PathRenderer(context, basicStroke, strokeColor, fillColor);
         }
         renderer.applyStroke(path);
      }
   }
   
   /**
    * Used to draw dash patterns on lines of pixels.
    * 
    * @param    {CanvasRenderingContext2D} context
    *           The canvas renderer.
    * @param    {Object} basicStroke
    *           The stroke style object.
    * @param    {Number[]} strokeColor
    *           The array of 3 integer values between 0 and 255 inclusive, representing
    *           a stroke color.
    * @param    {Number[]} fillColor
    *           The array of 3 integer values between 0 and 255 inclusive, representing
    *           a fill color.
    */
   function PathRenderer(ctx, basicStroke, strokeColor, fillColor)
   {
      var pixel = ctx.createImageData(1, 1);
      var pixelData = pixel.data;
      pixelData[3] = 0xFF;

      /**
       * Normalizes a dash offset distance according to the given dash pattern.
       * A normalized dash offset distance is a distance that is cut the corresponding segment
       * from the dash pattern that is extended periodically in order to cover the target dash
       * offset.
       * 
       * @param    {Array} dash
       *           The dash array that specifies the number of units to be drawn followed
       *           by the number of units to be skipped.
       * @param    {Number} dashOffset
       *           The offset distance from the first pattern point to the point which becomes
       *           a start point of the dash pattern.
       * 
       * @return   The container {startPhase : phase, startDashOn : dashOn, startIdx : idx},
       *           where the startPhase field holds a normalized dash offset, startDashOn defines
       *           to draw or to skip, the value of startIdx is the current dash pattern index.
       */
      function normalize(dash, dashOffset)
      {
         var phase = dashOffset;
         var idx = 0;
         var dashOn = true;
         var d;
         while (phase >= (d = dash[idx]))
         {
             phase -= d;
             idx = (idx + 1) % dash.length;
             dashOn = !dashOn;
         }

         return {startPhase : phase, startDashOn : dashOn, startIdx : idx};
      }
      
      /**
       * Apply the given stroke to the line of pixels called a path.
       * 
       * @param    {Array} path
       *           The array of points that forms the outline of the target area.
       */
      this.applyStroke = function (path)
      {
         var dash = basicStroke.getDash();
         if (!dash || dash.length == 0)
         {
            return;
         }
         var dashOffset = basicStroke.getDashPhase();
         var len = dash.length;
         
         var patternObj = normalize(dash, dashOffset);
         
         var dashOn = patternObj.startDashOn;
         var phase  = patternObj.startPhase;
         var idx    = patternObj.startIdx;
         var rest = dash[idx] - phase;
         for (var i = 0; i < path.length; i++)
         {
            var point = path[i];
            if (rest < 1)
            {
               idx = (idx + 1) % len;
               rest = dash[idx];
               dashOn = !dashOn;
            }
            rest -= 1;

            if (dashOn)
            {
               pixelData[0] = strokeColor[0];
               pixelData[1] = strokeColor[1];
               pixelData[2] = strokeColor[2];
            }
            else
            {
               pixelData[0] = fillColor[0];
               pixelData[1] = fillColor[1];
               pixelData[2] = fillColor[2];
            }
            ctx.putImageData(pixel, point.x, point.y);
         }

      }
   }
   
   /**
    * Used to widen lines of pixels.
    * 
    * @param    {CanvasRenderingContext2D} context
    *           The canvas renderer.
    * @param    {Object} basicStroke
    *           The stroke style object.
    * @param    {Number[]} strokeColor
    *           The array of 3 integer values between 0 and 255 inclusive, representing
    *           a stroke color.
    * @param    {Number[]} fillColor
    *           The array of 3 integer values between 0 and 255 inclusive, representing
    *           a fill color.
    */
   function WidenPathRenderer(ctx, basicStroke, strokeColor, fillColor)
   {
      var width = basicStroke.getWidth();
      var image = ctx.createImageData(width, width);
      var imageData = image.data;
      
      for (var i = 0; i < 4 * width * width; i += 4)
      {
         imageData[i] = strokeColor[0];
         imageData[i + 1] = strokeColor[1];
         imageData[i + 2] = strokeColor[2];
         imageData[i + 3] = 0xFF;
      }

      /**
       * To widen the line of pixels called a path according to the stroke style.
       * 
       * @param    {Array} path
       *           The array of points that forms the outline of the target area.
       */
      this.applyStroke = function (path)
      {
         for (var i = 0; i < path.length; i++)
         {
            var point = path[i];
            ctx.putImageData(image, point.x - (width >> 1), point.y - (width >> 1));
         }
      }
   }
}